#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import html


def fixture_function_str():
    return str("test")


html_escape = html.escape
