TELEMETRY_METRIC_TYPE_COUNT = "count"
TELEMETRY_METRIC_TYPE_GAUGE = "gauge"
TELEMETRY_METRIC_TYPE_RATE = "rate"
TELEMETRY_METRIC_TYPE_DISTRIBUTIONS = "distributions"

TELEMETRY_NAMESPACE_TAG_TRACER = "tracers"
TELEMETRY_NAMESPACE_TAG_APPSEC = "appsec"

TELEMETRY_TYPE_GENERATE_METRICS = "generate-metrics"
TELEMETRY_TYPE_DISTRIBUTION = "distributions"
